pbar_colour = '#509faf'
pbar_message = {'calibrator': 'Calibrating data',
                'colour_eq': 'Applying colour equation',
                'converter': 'Converting data',
                'correction': 'Correcting systems',
                'photometry': 'Generating photometry'}
pbar_units = {'calibrator': 'spec',
              'colour_eq': 'syst',
              'converter': 'spec',
              'correction': 'syst',
              'photometry': 'spec'}
