"""
nature.py
====================================
Module to hold "natural" constants.
"""

# Constants extracted from Gaia Parameter Database (Nature section)
C = 2.99792458E8
PLANCK = 6.62607004E-34
